from datetime import datetime
from typing import Any, Callable

from pydantic import GetJsonSchemaHandler, PlainSerializer, WithJsonSchema
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema
from typing_extensions import Annotated

from pytoncenter.address import Address

PyDatetime = Annotated[
    datetime,
    PlainSerializer(lambda x: int(x.timestamp()), return_type=int),
    WithJsonSchema({"type": "number"}, mode="serialization"),
]


class _AddressTypeAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        """
        - str will be parsed as Address
        - Address type will be parsed as Address without any changes
        - Nothing else will pass validation
        - Serialization will always return just an str
        """

        def validate_from_str(value: str) -> Address:
            return Address(value)

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(Address),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda x: x.to_string(
                    is_user_friendly=True,
                    is_url_safe=True,
                    is_test_only=False,
                )
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler) -> JsonSchemaValue:
        return handler(core_schema.str_schema())


AddressLike = Annotated[str, Address, _AddressTypeAnnotation]



from typing import Literal, Optional, overload

from pytoncenter.v2.api import AsyncTonCenterClientV2
from pytoncenter.v3.api import AsyncTonCenterClientV3


@overload
def get_client(version: Literal["v2"], network: Literal["mainnet", "testnet"], qps: Optional[float] = None, *args, **kwargs) -> AsyncTonCenterClientV2: ...


@overload
def get_client(version: Literal["v3"], network: Literal["mainnet", "testnet"], qps: Optional[float] = None, *args, **kwargs) -> AsyncTonCenterClientV3: ...


def get_client(version: Literal["v2", "v3"], network: Literal["mainnet", "testnet"], *args, **kwargs):
    if version == "v2":
        return AsyncTonCenterClientV2(network=network, *args, **kwargs)
    if version == "v3":
        return AsyncTonCenterClientV3(network=network, *args, **kwargs)
    raise ValueError(f"Invalid version {version}")


__version__ = "0.0.12"



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, List, OrderedDict, TypedDict, Union

from tonpy import CellSlice

from pytoncenter.v2.types import GetMethodResult
from pytoncenter.v3.models import (
    GetMethodParameterOutput,
    GetMethodParameterType,
    RunGetMethodResponse,
)

from .address import Address
from .utils import decode_base64

__all__ = ["BaseType", "Types", "BaseDecoder", "Decoder", "JettonDataDecoder"]


GetMethodResultType = Union[GetMethodResult, RunGetMethodResponse]


class BaseType:
    type: GetMethodParameterType

    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def decode(self, data: Any):
        raise NotImplementedError


class Types:
    class Raw(BaseType):
        type = "unsupported_type"

        def decode(self, data: Any):
            return data

    class Number(BaseType):
        type = "num"

        def decode(self, hex_str: str) -> int:
            return int(hex_str, 16)

    class Address(BaseType):
        type = "cell"

        def decode(self, cell: str) -> Address:
            return Address(CellSlice(cell).load_address())

    class Bool(BaseType):
        type = "num"

        def decode(self, hex_str: str) -> bool:
            return bool(int(hex_str, 16))

    class Cell(BaseType):
        type = "cell"

        def decode(self, cell: str) -> CellSlice:
            return CellSlice(cell)

    class B64String(BaseType):
        type = "cell"

        def decode(self, cell: str) -> str:
            return decode_base64(cell)

    class List(BaseType):
        type = "list"

        def __init__(self, name: str, *typs: BaseType) -> None:
            super().__init__(name)
            self.typs = typs

        def decode(self, data: Any):
            warnings.warn("List type is not supported yet, return raw data", RuntimeWarning)
            return data

    class Tuple(BaseType):
        type = "tuple"

        def __init__(self, name: str, *typs: BaseType) -> None:
            super().__init__(name)
            self.typs = typs

        def decode(self, data: Any):
            warnings.warn("Tuple type is not supported yet, return raw data", RuntimeWarning)
            return data

    class Slice(BaseType):
        type = "slice"

        def decode(self, data: Any):
            warnings.warn("Slice type is not supported yet, return raw data", RuntimeWarning)
            return data


class BaseDecoder(ABC):
    """
    BaseDecoderDecoder decodes the result of the return of the TON smart contract get method
    """

    @abstractmethod
    def decode(self, data: GetMethodResultType):
        raise NotImplementedError


class Decoder(BaseDecoder):
    def __init__(self, *typ: BaseType) -> None:
        self.types = typ
        assert len(self.types) == len(set(field.name for field in self.types)), "Field names must be unique"

    def _validate(self, data: GetMethodResultType) -> List[GetMethodParameterOutput]:
        if isinstance(data, RunGetMethodResponse):
            assert len(self.types) == len(data.stack), "Fields count must be equal to data count"
            return data.stack

        assert len(self.types) == len(data), "Fields count must be equal to data count"
        assert isinstance(data, list), "Data must be a list"
        if all(isinstance(data[i], dict) for i in range(len(data))):
            new_data = []
            for i in range(len(data)):
                assert self.types[i] is None or self.types[i].type == data[i]["type"], f"Field {i} type must be {self.types[i].type}, but got {data[i]['type']}"
                _type = data[i].get("type")
                _value = data[i].get("value")
                if _type == "cell":
                    new_data.append(GetMethodParameterOutput(type=_type, value=_value.get("bytes", "")))
                else:
                    new_data.append(GetMethodParameterOutput(type=_type, value=_value))
            return new_data
        raise ValueError("Data must be a RunGetMethodResponse(v3) or a list of GetMethodResult(v2)")

    def decode(self, data: GetMethodResultType) -> Dict[str, Any]:
        _data = self._validate(data)
        return OrderedDict((field.name, field.decode(_data[i].value)) for i, field in enumerate(self.types))


class AutoDecoder(BaseDecoder):
    def _transform(self, data: GetMethodResultType) -> List[GetMethodParameterOutput]:
        if isinstance(data, RunGetMethodResponse):
            return data.stack
        assert isinstance(data, list), "Data must be a list"
        if all(isinstance(data[i], dict) for i in range(len(data))):
            new_data = []
            for i in range(len(data)):
                type = data[i].get("type")
                value = data[i].get("value")
                if type == "cell":
                    new_data.append(GetMethodParameterOutput(type=type, value=value.get("bytes", "")))
                else:
                    new_data.append(GetMethodParameterOutput(type=type, value=value))
            return new_data
        raise ValueError("Data must be a RunGetMethodResponse(v3) or a list of GetMethodResult(v2)")

    def decode(self, data: GetMethodResultType) -> Dict[str, Any]:
        _data = self._transform(data)
        output = {}

        def _recursive_decode(data: GetMethodParameterOutput, prefix: str = "idx_", idx: int = 0, depth: int = 0):
            field_name = f"{prefix}{idx}"

            if data.type == "cell":
                assert isinstance(data.value, str), "Cell value must be a string"
                try:
                    t = Types.Address(field_name)
                    return t.name, t.decode(data.value)
                except:
                    t = Types.B64String(field_name)
                    return t.name, t.decode(data.value)
            if data.type == "slice":
                assert isinstance(data.value, list), "Slice value must be a list"
                t = Types.Slice(field_name)
                return t.name, t.decode(data.value)
            if data.type == "list" or data.type == "tuple":
                assert isinstance(data.value, list), "List value must be a list"
                results = []
                for i, t in enumerate(data.value):
                    result, _ = _recursive_decode(t, f"{field_name}_", i, depth + 1)
                    results.append(result)
                return field_name, results
            elif data.type == "num":
                assert isinstance(data.value, str), "Number value must be a string"
                t = Types.Number(field_name)
                return t.name, t.decode(data.value)
            else:
                t = Types.Raw(field_name)
                return t.name, t.decode(data.value)

        for idx, field in enumerate(_data):
            name, value = _recursive_decode(field, idx=idx)
            output[name] = value

        return output


JettonDataDict = TypedDict(
    "JettonDataDict",
    {
        "total_supply": int,
        "mintable": bool,
        "admin_address": Address,
        "jetton_content": str,
        "jetton_wallet_code": str,
    },
)


class JettonDataDecoder(BaseDecoder):
    """
    JettonDataDecoder is a singleton class that decodes the result of the get_jetton_data method
    """

    decoder = Decoder(
        Types.Number("total_supply"),
        Types.Bool("mintable"),
        Types.Address("admin_address"),
        Types.Cell("jetton_content"),
        Types.B64String("jetton_wallet_code"),
    )
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(JettonDataDecoder, cls).__new__(cls)
        return cls._instance

    def decode(self, data: GetMethodResultType) -> JettonDataDict:
        return self.decoder.decode(data)



from typing import List, Literal, TypedDict, Union, Dict, Any

AccountAddress = TypedDict("AccountAddress", {"@type": Literal["accountAddress"], "account_address": str})
AccountState = TypedDict("AccountState", {"@type": Literal["raw.accountState"], "code": str, "data": str, "frozen_hash": str})
BlockID = TypedDict("BlockID", {"@type": Literal["ton.blockIdExt"], "workchain": int, "shard": str, "seqno": int, "root_hash": str, "file_hash": str})
TransactionID = TypedDict("TransactionID", {"@type": Literal["'internal.transactionId'"], "lt": int, "hash": str})
WalletInformation = TypedDict("WalletInformation", {"wallet": bool, "balance": int, "account_state": str, "last_transaction_id": TransactionID})
AddressInformation = TypedDict(
    "AddressInformation",
    {
        "@type": Literal["raw.fullAccountState"],
        "balance": int,
        "code": str,
        "data": str,
        "last_transaction_id": TransactionID,
        "block_id": BlockID,
        "frozen_hash": str,
        "sync_utime": int,
        "@extra": str,
        "state": str,
    },
)
ExtentedAddressInformation = TypedDict(
    "ExtentedAddressInformation",
    {
        "@type": Literal["fullAccountState"],
        "address": AccountAddress,
        "balance": str,
        "last_transaction_id": TransactionID,
        "block_id": BlockID,
        "sync_utime": int,
        "account_state": AccountState,
        "revision": int,
        "@extra": str,
    },
)
MsgRawDataWithBody = TypedDict("MsgRawDataWithBody", {"@type": Literal["msg.dataRaw"], "body": str, "init_state": str})
MsgRawDataWithComment = TypedDict("MsgRawDataWithComment", {"@type": Literal["msg.dataText"], "text": str})
Message = TypedDict(
    "Message",
    {
        "@type": Literal["raw.message"],
        "source": str,
        "destination": str,
        "value": str,
        "foward_fee": str,
        "ihr_fee": str,
        "created_lt": str,
        "body_hash": str,
        "msg_data": Union[MsgRawDataWithBody, MsgRawDataWithComment],
        "messsage": str,
        "init_state": str,
    },
)
Tx = TypedDict(
    "Tx",
    {
        "@type": Literal["raw.transaction"],
        "address": AccountAddress,
        "utime": int,
        "data": str,
        "transaction_id": TransactionID,
        "fee": str,
        "storage_fee": str,
        "other_fee": str,
        "in_msg": Message,
        "out_msgs": List[Message],
    },
)
MasterChainInfo = TypedDict("MasterChainInfo", {"last": BlockID, "state_root_hash": str, "init": BlockID, "@extra": str})
ConsensusBlock = TypedDict("ConsensusBlock", {"consensus_block": int, "timestamp": float})
Signature = TypedDict("Signature", {"@type": Literal["blocks.signature"], "node_id_short": str, "signature": str})
BlockSignatures = TypedDict("BlockSignatures", {"@type": Literal["blocks.blockSignatures"], "id": BlockID, "signatures": List[Signature], "@extra": str})
Shards = TypedDict("Shards", {"@type": Literal["blocks.shard"], "shards": List[BlockID], "@extra": str})
BlockTxs = TypedDict("BlockTxs", {"@extra": str, "@type": Literal["blocks.transactions"], "id": BlockID, "incomplete": bool, "req_count": int, "transactions": List[Tx]})
JettonContentPayload = TypedDict("JettonContentPayload", {"image": str, "name": str, "symbol": str, "decimals": int, "description": str})
JettonContent = TypedDict("JettonContent", {"type": str, "data": JettonContentPayload})
JettonMasterData = TypedDict(
    "JettonMasterData", {"total_supply": int, "mintable": bool, "admin_address": str, "jetton_content": JettonContent, "jetton_wallet_code": str, "contract_type": Literal["jetton_master"]}
)
JettonWalletData = TypedDict("JettonWalletData", {"balance": int, "owner": str, "jetton": str, "jetton_wallet_code": str, "contract_type": Literal["jetton_wallet"]})
NFTCollectionContent = TypedDict("NFTCollectionContent", {"type": str, "data": str})
NFTCollectionData = TypedDict("NFTCollectionData", {"next_item_index": int, "owner_address": str, "contract_type": Literal["nft_collection"], "collection_content": NFTCollectionContent})
NFTItemData = TypedDict("NFTItemData", {"init": bool, "index": int, "owner_address": str, "collection_address": str, "contract_type": Literal["nft_item"], "content": NFTCollectionContent})
BounceableAddress = TypedDict("BounceableAddress", {"b64": str, "b64url": str})
DetectAddressResult = TypedDict("DetectAddressResult", {"raw_form": str, "bounceable": BounceableAddress, "non_bounceable": BounceableAddress, "given_type": str, "test_only": bool})
Fees = TypedDict("Fees", {"@type": Literal["fees"], "in_fws_fee": str, "storage_fee": str, "gas_fee": str, "fwd_fee": str})
EstimateResult = TypedDict("EstimateResult", {"@type": Literal["query.fees"], "source_fees": Fees, "destination_fees": List[Fees], "@extra": str})
TraceTx = TypedDict(
    "TraceTx",
    {
        "@type": Literal["raw.transaction"],
        "address": AccountAddress,
        "utime": int,
        "data": str,
        "transaction_id": TransactionID,
        "fee": str,
        "storage_fee": str,
        "other_fee": str,
        "in_msg": Message,
        "children": List["TraceTx"],
    },
)
GetMethodResult = List[Dict[Literal["type", "value"], Union[str, Dict[str, Any]]]]



from __future__ import annotations

from typing import List, Literal, Optional

from pydantic import BaseModel, Field, model_validator

from .types import AddressLike, PyDatetime

__all__ = [
    "GetBlockRequest",
    "GetMasterchainBlockShardsRequest",
    "GetTransactionsRequest",
    "GetTransactionByMasterchainBlockRequest",
    "GetTransactionByMessageRequest",
    "GetAdjacentTransactionsRequest",
    "GetTracesRequest",
    "GetTransactionTraceRequest",
    "GetMessagesRequest",
    "GetNFTCollectionsRequest",
    "GetNFTItemsRequest",
    "GetNFTTransfersRequest",
    "GetJettonMastersRequest",
    "GetJettonWalletsRequest",
    "GetJettonTransfersRequest",
    "GetJettonBurnsRequest",
    "GetTopAccountsByBalanceRequest",
    "GetAccountRequest",
    "GetWalletRequest",
    "GetSpecifiedJettonWalletRequest",
    "GetSpecifiedNFTItemRequest",
    "GetMessageByHashRequest",
    "GetTransactionByHashRequest",
    "GetSourceTransactionRequest",
    "SubscribeTransactionRequest",
    "WaitMessageExistsRequest",
]


class GetBlockRequest(BaseModel):
    workchain: Optional[int] = Field(default=None, description="Block workchain")
    shard: Optional[str] = Field(default=None, description="Block shard id. Must be sent with workchain")
    seqno: Optional[int] = Field(default=None, description="Block block seqno. Must be sent with workchain and shard")
    start_utime: Optional[PyDatetime] = Field(default=None, description="Query blocks with generation UTC timestamp after given timestamp")
    end_utime: Optional[PyDatetime] = Field(default=None, description="Query blocks with generation UTC timestamp before given timestamp")
    start_lt: Optional[int] = Field(default=None, description="Query blocks with lt >= start_lt")
    end_lt: Optional[int] = Field(default=None, description="Query blocks with lt <= end_lt")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read.")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read.")
    sort: Literal["asc", "desc"] = Field(default="desc", description="Sort results by UTC timestamp")

    @model_validator(mode="after")
    def check_times(cls, values: GetBlockRequest):
        start_utime, end_utime = values.start_utime, values.end_utime
        if start_utime and end_utime and start_utime > end_utime:
            raise ValueError("\033[93mstart_utime must be earlier than end_utime\033[0m")
        return values

    @model_validator(mode="after")
    def checl_lts(cls, values: GetBlockRequest):
        start_lt, end_lt = values.start_lt, values.end_lt
        if start_lt and end_lt and start_lt > end_lt:
            raise ValueError("\033[93mstart_lt must be earlier than end_lt\033[0m")
        return values

    @model_validator(mode="after")
    def check_shard_and_seqno(cls, values: GetBlockRequest):
        shard, workchain, seqno = values.shard, values.workchain, values.seqno
        if shard and not workchain:
            raise ValueError("\033[93mworkchain is required if shard is specified\033[0m")
        if seqno and not (workchain and shard):
            raise ValueError("\033[93mworkchain and shard are required if seqno is specified\033[0m")
        return values


class GetMasterchainBlockShardsRequest(BaseModel):
    seqno: int = Field(description="Masterchain block seqno")
    include_mc_block: bool = Field(False, description="Include masterchain block")


class GetTransactionByHashRequest(BaseModel):
    hash: str = Field(description="Transaction hash. Acceptable in hex, base64 and base64url forms")


class GetTransactionsRequest(BaseModel):
    workchain: Optional[int] = Field(default=None, description="Block workchain")
    shard: Optional[str] = Field(default=None, description="Block shard id. Must be sent with workchain")
    seqno: Optional[int] = Field(default=None, description="Block block seqno. Must be sent with workchain and shard")
    account: Optional[AddressLike] = Field(default=None, description="The account address to get transactions. Can be sent in hex, base64 or base64url form")
    exclude_account: Optional[AddressLike] = Field(default=None, description="Exclude transactions on specified account addresses")
    hash: Optional[str] = Field(default=None, description="Transaction hash. Acceptable in hex, base64 and base64url forms")
    lt: Optional[int] = Field(default=None, description="Transaction logical time")
    start_utime: Optional[PyDatetime] = Field(default=None, description="Query blocks with generation UTC timestamp after given timestamp")
    end_utime: Optional[PyDatetime] = Field(default=None, description="Query blocks with generation UTC timestamp before given timestamp")
    start_lt: Optional[int] = Field(default=None, description="Query blocks with lt >= start_lt")
    end_lt: Optional[int] = Field(default=None, description="Query blocks with lt <= end_lt")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read.")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read.")
    sort: Literal["asc", "desc"] = Field(default="desc", description="Sort results by UTC timestamp")

    @model_validator(mode="after")
    def check_times(cls, values: GetBlockRequest):
        start_utime, end_utime = values.start_utime, values.end_utime
        if start_utime and end_utime and start_utime > end_utime:
            raise ValueError("\033[93start_utime must be earlier than end_utime\033[0m")
        return values

    @model_validator(mode="after")
    def checl_lts(cls, values: GetBlockRequest):
        start_lt, end_lt = values.start_lt, values.end_lt
        if start_lt and end_lt and start_lt > end_lt:
            raise ValueError("\033[93mstart_lt must be earlier than end_lt\033[0m")
        return values

    @model_validator(mode="after")
    def check_shard_and_seqno(cls, values: GetBlockRequest):
        shard, workchain, seqno = values.shard, values.workchain, values.seqno
        if shard and not workchain:
            raise ValueError("\033[93mworkchain is required if shard is specified\033[0m")
        if seqno and not (workchain and shard):
            raise ValueError("\033[93mworkchain and shard are required if seqno is specified\033[0m")
        return values


class GetTransactionByMasterchainBlockRequest(BaseModel):
    seqno: int = Field(description="Masterchain block seqno")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read.")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read.")
    sort: Literal["asc", "desc"] = Field(default="desc", description="Sort results by UTC timestamp")


class GetTransactionByMessageRequest(BaseModel):
    direction: Literal["in", "out"] = Field(description="Message direction")
    msg_hash: str = Field(description="Message hash. Acceptable in hex, base64 and base64url forms")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read.")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read.")


class GetSourceTransactionRequest(BaseModel):
    hash: str = Field(description="Transaction hash. Acceptable in hex, base64 and base64url forms")


class GetAdjacentTransactionsRequest(BaseModel):
    hash: str = Field(description="Transaction hash. Acceptable in hex, base64 and base64url forms")
    direction: Literal["in", "out", "both"] = Field(default="both", description="Message direction")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read.")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read.")
    sort: Literal["asc", "desc", "none"] = Field(default="desc", description="Sort results by UTC timestamp")
    full: bool = Field(default=False, description="Automatically paginate through the entire history of the transaction", exclude=True)


class GetTracesRequest(BaseModel):
    tx_hash: List[str] = Field(default=[], description="List of transaction hashes")
    trace_id: List[str] = Field(default=[], description="Trace id")

    @model_validator(mode="after")
    def check_txhash_traceid(cls, values: GetTracesRequest):
        tx_hash, trace_id = values.tx_hash, values.trace_id
        if (not tx_hash and not trace_id) or (tx_hash and trace_id):
            raise ValueError("\033[93mExact one parameter should be used\033[0m")
        return values


class GetTransactionTraceRequest(BaseModel):
    hash: str = Field(description="Transaction hash. Acceptable in hex, base64 and base64url forms")
    sort: Literal["none", "asc", "desc"] = Field(default="asc", description="Sort transactions by lt")


class GetMessageByHashRequest(BaseModel):
    hash: str = Field(description="Message hash. Acceptable in hex, base64 and base64url forms")


class GetMessagesRequest(BaseModel):
    hash: Optional[str] = Field(default=None, description="Message hash. Acceptable in hex, base64 and base64url forms")
    source: Optional[AddressLike] = Field(default=None, description="The source account address. Can be sent in hex, base64 or base64url form")
    destination: Optional[AddressLike] = Field(default=None, description="The destination account address. Can be sent in hex, base64 or base64url form")
    body_hash: Optional[str] = Field(default=None, description="Message body hash. Acceptable in hex, base64 and base64url forms")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")


class GetNFTCollectionsRequest(BaseModel):
    collection_address: Optional[AddressLike] = Field(default=None, description="NFT collection address. Must be sent in hex, base64 and base64url forms")
    owner_address: Optional[AddressLike] = Field(default=None, description="NFT owner address. Must be sent in hex, base64 and base64url forms")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")


class GetNFTItemsRequest(BaseModel):
    address: Optional[AddressLike] = Field(default=None, description="NFT address. Must be sent in hex, base64 and base64url forms.")
    owner_address: Optional[AddressLike] = Field(default=None, description="Address of NFT owner. Must be sent in hex, base64 and base64url forms")
    collection_address: Optional[AddressLike] = Field(default=None, description="NFT collection address. Must be sent in hex, base64 and base64url forms")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")


class GetSpecifiedNFTItemRequest(BaseModel):
    owner_address: AddressLike = Field(..., description="Address of NFT owner. Must be sent in hex, base64 and base64url forms")
    collection_address: AddressLike = Field(..., description="NFT collection address. Must be sent in hex, base64 and base64url forms")


class GetNFTTransfersRequest(BaseModel):
    address: Optional[AddressLike] = Field(default=None, description="Address of NFT owner. Must be sent in hex, base64 and base64url forms")
    item_address: Optional[AddressLike] = Field(default=None, description="NFT item address. Must be sent in hex, base64 and base64url forms")
    collection_address: Optional[AddressLike] = Field(default=None, description="NFT collection address. Must be sent in hex, base64 and base64url forms")
    direction: Literal["in", "out", "both"] = Field(default="both", description="Direction transactions by lt")
    start_utime: Optional[PyDatetime] = Field(default=None, description="Query transactions with generation UTC timestamp after given timestamp")
    end_utime: Optional[PyDatetime] = Field(default=None, description="Query transactions with generation UTC timestamp before given timestamp")
    start_lt: Optional[int] = Field(default=None, description="Query transactions with lt >= start_lt")
    end_lt: Optional[int] = Field(default=None, description="Query transactions with lt <= end_lt")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")
    sort: Literal["asc", "desc"] = Field(default="desc", description="Sort results by UTC timestamp")

    @model_validator(mode="after")
    def check_times(cls, values: GetNFTTransfersRequest):
        start_utime, end_utime = values.start_utime, values.end_utime
        if start_utime and end_utime and start_utime > end_utime:
            raise ValueError("\033[93mstart_utime must be earlier than end_utime\033[0m")
        return values

    @model_validator(mode="after")
    def checl_lts(cls, values: GetBlockRequest):
        start_lt, end_lt = values.start_lt, values.end_lt
        if start_lt and end_lt and start_lt > end_lt:
            raise ValueError("\033[93mstart_lt must be earlier than end_lt\033[0m")
        return values


class GetJettonMastersRequest(BaseModel):
    address: Optional[AddressLike] = Field(default=None, description="Jetton Master address. Must be sent in hex, base64 and base64url forms")
    admin_address: Optional[AddressLike] = Field(default=None, description="Address of Jetton Master's admin. Must be sent in hex, base64 and base64url forms")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")


class GetJettonWalletsRequest(BaseModel):
    address: Optional[AddressLike] = Field(default=None, description="Jetton wallet address. Must be sent in hex, base64 and base64url forms")
    owner_address: Optional[AddressLike] = Field(default=None, description="Address of Jetton wallet's owner. Must be sent in hex, base64 and base64url forms")
    jetton_address: Optional[AddressLike] = Field(default=None, description="Jetton Master. Must be sent in hex, base64 and base64url forms")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")


class GetSpecifiedJettonWalletRequest(BaseModel):
    owner_address: AddressLike = Field(..., description="Address of Jetton wallet's owner. Must be sent in hex, base64 and base64url forms")
    jetton_address: AddressLike = Field(..., description="Jetton Master. Must be sent in hex, base64 and base64url forms")


class JettonFilter(BaseModel):
    address: Optional[AddressLike] = Field(default=None, description="Account address. Must be sent in hex, base64 and base64url forms")
    jetton_wallet: Optional[AddressLike] = Field(default=None, description="Jetton wallet address. Must be sent in hex, base64 and base64url forms")
    jetton_master: Optional[AddressLike] = Field(default=None, description="Jetton master address. Must be sent in hex, base64 and base64url forms")
    direction: Literal["in", "out", "both"] = Field(default="both", description="Direction transactions by lt")
    start_utime: Optional[PyDatetime] = Field(default=None, description="Query transactions with generation UTC timestamp after given timestamp")
    end_utime: Optional[PyDatetime] = Field(default=None, description="Query transactions with generation UTC timestamp before given timestamp")
    start_lt: Optional[int] = Field(default=None, description="Query transactions with lt >= start_lt")
    end_lt: Optional[int] = Field(default=None, description="Query transactions with lt <= end_lt")
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")
    sort: Literal["asc", "desc"] = Field(default="desc", description="Sort results by UTC timestamp")

    @model_validator(mode="after")
    def check_times(cls, values: GetNFTTransfersRequest):
        start_utime, end_utime = values.start_utime, values.end_utime
        if start_utime and end_utime and start_utime > end_utime:
            raise ValueError("\033[93mstart_utime must be earlier than end_utime\033[0m")
        return values

    @model_validator(mode="after")
    def checl_lts(cls, values: GetBlockRequest):
        start_lt, end_lt = values.start_lt, values.end_lt
        if start_lt and end_lt and start_lt > end_lt:
            raise ValueError("\033[93mstart_lt must be earlier than end_lt\033[0m")
        return


class SubscribeTransactionRequest(BaseModel):
    account: AddressLike = Field(..., description="Account address. Must be sent in hex, base64 and base64url forms")
    start_time: Optional[PyDatetime] = Field(default=None, description="Query transactions with generation UTC timestamp after given timestamp")
    interval: float = Field(default=2.0, description="Interval in seconds to check for new transactions")
    limit: int = Field(default=256, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")


class WaitMessageExistsRequest(BaseModel):
    msg_hash: str = Field(..., description="Message hash. Acceptable in hex, base64 and base64url forms")
    max_retry: Optional[int] = Field(default=None, description="Max retry times to check for new messages, None for infinite")
    interval: float = Field(default=2.0, description="Interval in seconds to check for new messages")


class GetJettonTransfersRequest(JettonFilter): ...


class GetJettonBurnsRequest(JettonFilter): ...


class GetTopAccountsByBalanceRequest(BaseModel):
    limit: int = Field(default=128, ge=1, le=256, description="Limit number of queried rows. Use with offset to batch read")
    offset: int = Field(default=0, ge=0, description="Skip first N rows. Use with limit to batch read")


class GetAccountRequest(BaseModel):
    address: AddressLike = Field(description="Account address. Account address. Can be sent in raw or user-friendly form")


class GetWalletRequest(BaseModel):
    address: AddressLike = Field(description="Account address. Account address. Can be sent in raw or user-friendly form")
